/*------------------------------------------------------------------*\
* Image manipulation routines
* 3D group (Danilo, Diego and Diogo)
\*------------------------------------------------------------------*/ 

#include <stdlib.h>
#include <assert.h>
#include <limits.h>
#include <string.h>

#include "image.h"

/*------------------------------------------------------------------*\
* Internal function prototypes
\*------------------------------------------------------------------*/ 
static int check_input(p_image in);
static int alloc_rgb(int w, int h, uchar **r, uchar **g, uchar **b);
static int splitrgb_rgb(const p_image in, uchar *r, uchar *g, uchar *b);
static int splitrgb_mono(const p_image in, uchar *r, uchar *g, uchar *b);

/*------------------------------------------------------------------*\
* Creates new white opaque image. Memory is allocated for the 
* image, not for the structure.
* Input:
*   t: type of image to be created
*   w, h: image dimensions
* Output:
* 	out: structure filled with image
* Returns:
*   0 in case of error, 1 if sucessfull
\*------------------------------------------------------------------*/ 
int image_create(int w, int h, e_imagetype t, p_image out)
{
  int r;
  assert(w >= 0 && h >= 0);
	if (w < 0 || h < 0) 
		return 0;

	assert(out);
	if (!out)
		return 0;

  r = w;
  if (r & 0x03) {
    r |= 0x03;
    r++;
  }

	out->buffer = (uchar *) malloc(r*h*t);
	if (!out->buffer)
		return 0;

	out->reference = 0;
	out->width = w;
	out->row = r;
	out->height = h;
	out->type = t;

	image_clear(out, 255);

	return 1;
}

/*------------------------------------------------------------------*\
* Crops image borders until non-black pixels are found.
* Input:
*   in: image to be cropped
* Returns:
*   0 in case of error, 1 if sucessfull
\*------------------------------------------------------------------*/ 
int image_compactmono(p_image in)
{
	int ymin = INT_MAX, ymax = INT_MIN, xmin = INT_MAX, xmax = INT_MIN;
	int border_x, border_y;
	int skip;
	int x, y;
	uchar *p;

	if (!check_input(in))
		return 0;

	skip = in->row - in->width;

	p = in->buffer;
	for (y = 0; y < in->height; y++) {
		for (x = 0; x < in->width; x++) {
			if (*p) {
				if (y < ymin)
					ymin = y;
				if (y > ymax)
					ymax = y;
				if (x > xmax)
					xmax = x;
				if (x < xmin)
					xmin = x;
			}
			p++;
		}
		p += skip;
	}

	if (ymin >= ymax || xmin >= xmax)
		return 0;

	ymax = in->height - ymax - 1;
	xmax = in->width - xmax - 1;

	border_x = xmin < xmax ? xmin : xmax;
	border_y = ymin < ymax ? ymin : ymax;

	in->buffer = in->buffer + border_y*in->row + border_x;
	in->width = in->width - 2*border_x;
	in->height = in->height - 2*border_y;

	return 1;
}

/*------------------------------------------------------------------*\
* Clears an image.
* Input:
*   in: image to be cleared
* Returns:
*   0 in case of error, 1 if sucessfull
\*------------------------------------------------------------------*/ 
int image_clear(p_image in, int c)
{
	int y;
	uchar *p;

	if (!check_input(in))
		return 0;

	p = in->buffer;
	for (y = 0; y < in->height; y++) {
		memset(p, c, in->width*in->type);
		p += in->row*in->type;
	}

	return 1;
}

/*------------------------------------------------------------------*\
* Frees an image if it is not a reference.
* Input:
*   in: image
* Returns:
*   0 in case of error, 1 if sucessfull
\*------------------------------------------------------------------*/ 
int image_kill(p_image in)
{
	if (!check_input(in))
		return 0;

	if (in->reference)
		return 0;

	if (in->buffer)
	  free(in->buffer);
	in->buffer = NULL;
	in->row = 0;
	in->reference = 0;
	in->width = 0;
	in->height = 0;
	in->type = 0;
	return 1;
}

/*------------------------------------------------------------------*\
* Blends a alpha image into another, using specified base color.
* Input:
*   in: image to receive blending
*   a: alpha image to be blended
*   rgb: base color
*   xc, yc: center of blending in input image
* Returns:
*   0 in case of error, 1 if sucessfull
\*------------------------------------------------------------------*/ 
int image_blendrgb(p_image in, const p_image stroke, const uchar *rgb,
    int xc, int yc)
{
	uchar *alpha, *input;
	long la, li;			/* line jumps for both images */
	int xi, yi; 			/* lower left corner in input image */
	int xa, ya;				/* corresponding corner in alpha image */
	int wa, ha;				/* area in alpha image to be blended */
	int x, y;
	int r = rgb[0], g = rgb[1], b = rgb[2];
	int i, a, ai;

	if (!check_input(in))
		return 0;

	wa = stroke->width;
	xa = 0;
	xi = xc - wa/2;
	if (xi < 0) {
		wa += xi;
		xa -= xi;
		xi = 0;
	}
	if (xi > in->width) 
		return 0;
	if (wa <= 0)
		return 0;
	if (xi + wa >= in->width)
		wa = in->width - xi;

	ha = stroke->height;
	ya = 0;
	yi = yc - ha/2;
	if (yi < 0) {
		ha += yi;
		ya -= yi;
		yi = 0;
	}
	if (yi > in->height) 
		return 0;
	if (ha <= 0)
		return 0;
	if (yi + ha >= in->height)
		ha = in->height - yi;

	la = stroke->row*stroke->type;
	li = in->row*in->type;

	input = in->buffer + li*yi + xi*in->type;
	alpha = stroke->buffer + la*ya + xa*stroke->type;
	
	for (y = 0; y < ha; y++) {
		for (x = 0; x < wa; x++) {
			a = *alpha++; ai = 255 - a;
			i = a*r + ai*(*input); i >>= 8;
			*input++ = (uchar) i;
			i = a*g + ai*(*input); i >>= 8;
			*input++ = (uchar) i;
			i = a*b + ai*(*input); i >>= 8;
			*input++ = (uchar) i;
		}
		input += li - wa * in->type;
		alpha += la - wa * stroke->type;
	}

	return 1;
}

/*------------------------------------------------------------------*\
* Allocates memory and split an image into it's RGB channels
* Input:
*   in: image
* Output:
* 	w, h: image dimensions
* 	r, g, b: image channels
* Returns:
*   0 in case of error, 1 if sucessfull
\*------------------------------------------------------------------*/ 
int image_splitnewrgb(const p_image in, uchar **r, uchar **g, uchar **b, 
    int *w, int *h)
{
	if (!check_input(in))
		return 0;

	if (!alloc_rgb(in->width, in->height, r, g, b))
		return 0;

	*w = in->width;
	*h = in->height;

	return image_splitrgb(in, *r, *g, *b);
}

/*------------------------------------------------------------------*\
* Allocates memory and split an image into it's alpha channel
* Input:
*   in: image
* Output:
* 	w, h: image dimensions
* 	r, g, b: image channels
* Returns:
*   0 in case of error, 1 if sucessfull
\*------------------------------------------------------------------*/ 
int image_splitnewmono(const p_image in, uchar **a, int *w, int *h)
{
	int y;
	uchar *input, *output;
	if (!check_input(in))
		return 0;
	*a = (uchar *) malloc(in->width*in->height);
	if (!(*a))
		return 0;
	output = *a;
	input = in->buffer;
	for (y = 0; y < in->height; y++) {
		memcpy(output, input, in->width);
		output += in->width;
		input += in->row;
	}
	*w = in->width;
	*h = in->height;
	return 1;
}

/*------------------------------------------------------------------*\
* Splits an image into it's RGB channels
* Input:
*   in: image
* 	r, g, b: allocated image channels
* Output:
* 	r, g, b: splitted image channels
* Returns:
*   0 in case of error, 1 if sucessfull
\*------------------------------------------------------------------*/ 
int image_splitrgb(const p_image in, uchar *r, uchar *g, uchar *b)
{
 	if (!check_input(in))
		return 0;

	switch (in->type) {
		case IMAGE_RGB:
			splitrgb_rgb(in, r, g, b); 
			break;
		case IMAGE_MONO:
			splitrgb_mono(in, r, g, b); 
			break;
		default:
			return 0;
	}

	return 1;
}

/*------------------------------------------------------------------*\
* Creates a copy of an image.
* Input:
*   in: image
* Output:
* 	out: copy of image
* Returns:
*   0 in case of error, 1 if sucessfull
\*------------------------------------------------------------------*/ 
int image_copy(const p_image in, p_image out)
{
	int y;
	uchar *input, *output;
	if (!check_input(in))
	 	return 0;
	if (!image_create(in->width, in->height, in->type, out))
		return 0;
	output = out->buffer;
	input = in->buffer;
	for (y = 0; y < in->height; y++) {
		memcpy(output, input, in->width*in->type);
		output += out->row*out->type;
		input += in->row*in->type;
	}
	return 1;
}

/*------------------------------------------------------------------*\
* Creates an image from a mono channell. Memory is allocated for
* the new image, not for structure.
* Input:
*   a: image channel
* 	w, h: image dimensions
* Output:
* 	out: interlaced image structure
* Returns:
*   0 in case of error, 1 if sucessfull
\*------------------------------------------------------------------*/ 
int image_mergenewmono(const uchar *a, int w, int h, p_image out)
{
	uchar *p;
	int y;

	assert(a);
	if (!a)
		return 0;

	if (!image_create(w, h, IMAGE_MONO, out))
		return 0;

	p = out->buffer;
	for (y = 0; y < h; y++) {
  	memcpy(p, a, w);
    a += w;
    p += out->row;
	}

  return 1;
}

/*------------------------------------------------------------------*\
* Merge channels into an interlaced image. Memory is allocated for
* interlaced image, not for structure.
* Input:
*   r, g, b: image channels
* 	w, h: image dimensions
* Output:
* 	out: interlaced image structure
* Returns:
*   0 in case of error, 1 if sucessfull
\*------------------------------------------------------------------*/ 
int image_mergenewrgb(const uchar *r, const uchar *g, const uchar *b, 
	int w, int h, p_image out)
{
	uchar *p;
	int x, y;
  long skip;

	assert(r && g && b);
	if (!r || !g || !b)
		return 0;

	if (!image_create(w, h, IMAGE_RGB, out))
		return 0;

	p = out->buffer;
  skip = (out->row - out->width)*out->type;
	for (y = 0; y < h; y++) {
		for (x = 0; x < w; x++) {
			*p++ = *r++;
			*p++ = *g++;
			*p++ = *b++;
		}
    p += skip;
	}
	return 1;
}

/*------------------------------------------------------------------*\
* Creates an image pointing to part of another image.
* Input:
*   in: source image
*   xc, yc: center of crop region
* 	w, h: crop region dimensions
* Output:
* 	out: croped reference
* Returns:
*   0 in case of error, 1 if sucessfull
\*------------------------------------------------------------------*/ 
int image_crop(const p_image in, int xc, int yc, int w, int h, 
	p_image out)
{
	int x0, y0;

	if (!check_input(in))
		return 0;

	/* adjust region to fit in source image */
	x0 = xc - w/2;
	y0 = yc - h/2;
	if (x0 < 0) {
		w += x0;
		x0 = 0;
	}
	if (x0 > in->width) x0 = in->width;
	if (y0 < 0) {
		h += y0;
		y0 = 0;
	}
	if (y0 > in->height) y0 = in->height;
	if (x0 + w > in->width) w = in->width - x0; 
	if (y0 + h > in->height) h = in->height - y0; 

	/* create reference */
	out->width = w;
	out->height = h;
	out->type = in->type;
	out->row = in->row;
	out->reference = 1;
	out->buffer = in->buffer + y0*in->row*in->type + x0*in->type;
	return 1;
}

/*------------------------------------------------------------------*\
* Internal functions
\*------------------------------------------------------------------*/ 
static int alloc_rgb(int w, int h, uchar **r, uchar **g, uchar **b)
{
	assert(w >= 0 && h >= 0);
	if (w < 0 || h < 0)
		return 0;
	
	assert(r && g && b);
	if (!r || !g || !b)
		return 0;

	*r = (uchar *) malloc(w*h);
	*g = (uchar *) malloc(w*h);
	*b = (uchar *) malloc(w*h);

	if (!*r || !*g || !*b) {
		if (*r) free(r); *r = NULL;
		if (*g) free(g); *g = NULL;
		if (*b) free(b); *b = NULL;
		return 0;
	}
	return 1;
}

static int splitrgb_rgb(const p_image in, uchar *r, uchar *g, uchar *b)
{
	int x, y;
	long skip = (in->row - in->width)*in->type;
	uchar *p = in->buffer;
	for (y = 0; y < in->height; y++) {
		for (x = 0; x < in->width; x++) {
			*b++ = *p++;
			*g++ = *p++;
			*r++ = *p++;
		}
		/* jump to next scan line */
		p += skip;
	}
	return 1;
}

static int splitrgb_mono(const p_image in, uchar *r, uchar *g, uchar *b)
{
	int x, y;
	uchar *p = in->buffer;
	long skip = (in->row - in->width)*in->type;
	for (y = 0; y < in->height; y++) {
		for (x = 0; x < in->width; x++) {
			*b++ = *p;
			*g++ = *p;
			*r++ = *p++;
		}
		/* jump to next scan line */
		p += skip;
	}
	return 1;
}

int check_input(p_image in)
{
	assert(in && in->buffer);
	if (!in || !in->buffer)
		return 0;

	return 1;
}
