/*------------------------------------------------------------------*\
* Image manipulation routines
* 3D group (Danilo, Diego and Diogo)
\*------------------------------------------------------------------*/
#ifndef _IMAGE_H_
#define _IMAGE_H_

/*------------------------------------------------------------------*\
* Exported declarations
\*------------------------------------------------------------------*/
/* imate type. mamata: constants are equal to pixel storage size */
typedef enum e_imagetype {
	IMAGE_MONO = 1,
	IMAGE_RGB  = 3
} e_imagetype;

/* hate writing "unsigned char" */
typedef unsigned char uchar;

/* image structure */
typedef struct t_image {
	e_imagetype type;		/* pixel format */
	int width, height;		/* dimensions */
	long row;				/* row length */
	int reference;			/* is the image a refence to other image? */
	uchar *buffer;			/* pixel storage */
} t_image;

typedef t_image *p_image;

/*------------------------------------------------------------------*\
* Creates new white opaque image. Memory is allocated for the 
* image, not for the structure.
* Input:
*   t: type of image to be created
*   w, h: image dimensions
* Output:
*   out: structure filled with image
* Returns:
*   0 in case of error, 1 if sucessfull
\*------------------------------------------------------------------*/
int image_create(int w, int h, e_imagetype type, p_image out);

/*------------------------------------------------------------------*\
* Creates an image pointing to part of another image.
* Input:
*   in: source image
*   xc, yc: center of crop region
*   w, h: crop region dimensions
* Output:
*   out: croped reference
* Returns:
*   0 in case of error, 1 if sucessfull
\*------------------------------------------------------------------*/
int image_crop(const p_image in, int x, int y, int w, int h, p_image out);

/*------------------------------------------------------------------*\
* Clears an image.
* Input:
*   in: image to be cleared
*   c: value to fill image with
* Returns:
*   0 in case of error, 1 if sucessfull
\*------------------------------------------------------------------*/ 
int image_clear(p_image in, int c);

/*------------------------------------------------------------------*\
* Crops image borders until non-black pixels are found.
* Input:
*   in: image to be cropped
* Returns:
*   0 in case of error, 1 if sucessfull
\*------------------------------------------------------------------*/ 
int image_compactmono(p_image in);

/*------------------------------------------------------------------*\
* Blends a alpha image into another, using specified base color.
* Input:
*   in: image to receive blending
*   a: alpha image to be blended
*   rgb: base color
*   xc, yc: center of blending in input image
* Returns:
*   0 in case of error, 1 if sucessfull
\*------------------------------------------------------------------*/
int image_blendrgb(p_image in, const p_image stroke, const uchar *rgb,
    int xc, int yc);

/*------------------------------------------------------------------*\
* Frees an image if it is not a reference.
* Input:
*   in: image
* Returns:
*   0 in case of error, 1 if sucessfull
\*------------------------------------------------------------------*/
int image_kill(p_image in);

/*------------------------------------------------------------------*\
* Splits an image into it's RGB channels
* Input:
*   in: image
* 	r, g, b: allocated image channels
* Output:
* 	r, g, b: splitted image channels
* Returns:
*   0 in case of error, 1 if sucessfull
\*------------------------------------------------------------------*/ 
int image_splitrgb(const p_image in, uchar *r, uchar *g, uchar *b);

/*------------------------------------------------------------------*\
* Allocates memory and split an image into it's RGB channels
* Input:
*   in: image
* Output:
*   w, h: image dimensions
*   r, g, b: image channels
* Returns:
*   0 in case of error, 1 if sucessfull
\*------------------------------------------------------------------*/
int image_splitnewrgb(const p_image in, uchar **r, uchar **g, uchar **b, 
	int *w, int *h);

/*------------------------------------------------------------------*\
* Allocates memory and split an image into it's alpha channel
* Input:
*   in: image
* Output:
* 	w, h: image dimensions
* 	r, g, b: image channels
* Returns:
*   0 in case of error, 1 if sucessfull
\*------------------------------------------------------------------*/ 
int image_splitnewmono(const p_image in, uchar **a, int *w, int *h);

/*------------------------------------------------------------------*\
* Creates a copy of an image.
* Input:
*   in: image
* Output:
* 	out: copy of image
* Returns:
*   0 in case of error, 1 if sucessfull
\*------------------------------------------------------------------*/ 
int image_copy(const p_image in, p_image out);

/*------------------------------------------------------------------*\
* Merge channels into an interlaced image. Memory is allocated for
* interlaced image, not for structure.
* Input:
*   r, g, b: image channels
*   w, h: image dimensions
* Output:
*   out: interlaced image structure
* Returns:
*   0 in case of error, 1 if sucessfull
\*------------------------------------------------------------------*/
int image_mergenewrgb(const uchar *r, const uchar *g, const uchar *b, 
	int w, int h, p_image out);

/*------------------------------------------------------------------*\
* Creates an image from a mono channell. Memory is allocated for
* the new image, not for structure.
* Input:
*   a: image channel
* 	w, h: image dimensions
* Output:
* 	out: interlaced image structure
* Returns:
*   0 in case of error, 1 if sucessfull
\*------------------------------------------------------------------*/ 
int image_mergenewmono(const uchar *a, int w, int h, p_image out);

#endif
