#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <math.h>
#include <im.h>

#include "image.h"
#include "painter.h"
#include "dither.h"
#include "bench.h"
#include "warp.h"
#include "list.h"
#include "gauge.h"

int load_rgb(char *file, p_image out)
{
    uchar *red, *green, *blue;
    int width, height, type, pal_size, format;
    if (imFileFormat(file, &format) != IM_ERR_NONE)
		return 0;
    imImageInfo(file, &width, &height, &type, &pal_size);
    red = malloc(width * height);
    green = malloc(width * height);
    blue = malloc(width * height);
	if (!red || !green || !blue) {
		if (red) free(red);
		if (green) free(green);
		if (blue) free(blue);
		return 0;
	}
    imLoadRGB(file, red, green, blue);
	image_mergenewrgb(red, green, blue, width, height, out);
	free(red); free(green); free(blue);
	return 1;
}

int save_rgb(p_image in, char *file)
{
    uchar *red, *green, *blue;
    int width, height;
	int ret = 0;
    if (!image_splitnewrgb(in, &red, &green, &blue, &width, &height))
		return 0;
    if(imSaveRGB(width, height, IM_RAS, red, green, blue, file) != IM_ERR_NONE)
		ret = 0;
	free(red); free(green); free(blue);
	return ret;
}

int save_mono(p_image in, char *file)
{
    uchar *alpha;
    int width, height;
	int ret = 0;
    if (!image_splitnewmono(in, &alpha, &width, &height))
		return 0;
    if (imSaveRGB(width, height, IM_RAS, alpha, alpha, alpha, 
		file) != IM_ERR_NONE)
		ret = 0;
	free(alpha);
	return ret;
}

int load_mono(char *file, p_image out)
{
    uchar *red, *green, *blue;
    int width, height, type, pal_size, format;
    if (imFileFormat(file, &format) != IM_ERR_NONE)
		return 0;
    imImageInfo(file, &width, &height, &type, &pal_size);
    red = malloc(width * height);
    green = malloc(width * height);
    blue = malloc(width * height);
	if (!red || !green || !blue) {
		if (red) free(red);
		if (green) free(green);
		if (blue) free(blue);
		return 0;
	}
    imLoadRGB(file, red, green, blue);
	image_mergenewmono(red, width, height, out);
	free(red); free(green); free(blue);
	return 1;
}

void print_help(void)
{
	printf("Panterly renders an image. Works with Sun raster files.\n");
	printf("  paint <input> <stroke> <output>\n\n");
}

void error(const char *str, ...)
{
    va_list ap;
    va_start(ap, str);
    vprintf(str, ap);
    printf("\n");
    va_end(ap);
    exit(1);
}

char *string(const char *format, ...)
{
	static char buffer[1025];
    va_list ap;
    va_start(ap, format);
    vsprintf(buffer,format, ap);
    va_end(ap);
	return buffer;
}

int getstrokes(p_image input, p_image area, p_list *list, int S, 
	int factor, int level)
{
	int mark;
	int count = 0;
	t_image I;
	uchar *buffer;
	t_image position, inputscaled, areascaled;
	int x, y;

	printf("Generating stroke positions image.\n");
	mark = bench_mark();
	warp_scale(area, area->width/factor, area->height/factor, &areascaled);
	image_copy(&areascaled, &position);
	save_mono(&areascaled, string("area-%d.ras", factor));
	dither_mono(&position, 4*S/sqrt(level), 2.0f/level);
	printf("Done in %fs.\n", bench_gettime(mark));
	save_mono(&position, string("points-%d.ras", factor));

	warp_scale(input, input->width/factor, input->height/factor, 
		&inputscaled);

	printf("Collecting strokes.\n");
	mark = bench_mark();
	buffer = position.buffer;	
	gauge_setmax(position.height*position.width);
	gauge_start();
	for (y = 0; y < position.height; y++) {
		for (x = 0; x < position.width; x++) {
			if (!(*buffer)) {
				int xc, yc;
				float w, l;
				double theta;
				uchar rgb[3];
				image_create(S, S, IMAGE_MONO, &I);
				painter_strokepar(&inputscaled, x, y, S, &xc, &yc, &theta, &w, &l, rgb, &I);
				list_prepend(list, factor*xc, factor*yc, factor*w, factor*l, (float) theta, rgb, level);
				count++;
			}
			buffer++;
		}
		gauge_increase(position.width);
		gauge_print();
	}
	gauge_end();
	printf("Done in %fs.\n", bench_gettime(mark));
	image_kill(&position);
	image_kill(&I);
	return count;
}

#define _DEFS 15
#define MAX(a,b) ((a)>(b)?(a):(b))
int main(int argc, char **argv)
{
	int S;
	t_image input, output, stroke, area;
	p_list list;
	int mark, count;
	int factor, size, level;

	if (argc < 4) {
		print_help();
		return 1;
	}

	if (!load_rgb(argv[1], &input)) 
		error("Error reading input file <%s>.\n", argv[1]);

	if (!load_mono(argv[2], &stroke)) 
		error("Error reading stroke file <%s>.\n", argv[2]);

	if (argc < 5 || !(S = atoi(argv[4]))) {
		S = _DEFS;
		printf("Using default value for stroke dimension: %d\n", _DEFS);
	}

	warp_open(MAX(stroke.width*2, input.width), MAX(stroke.height*2, 
		input.height));
    bench_open();

	printf("Generating stroke area image.\n");
	mark = bench_mark();
	painter_strokearea(&input, S, &area);
	printf("Done in %fs.\n", bench_gettime(mark));
	save_mono(&area, string("area-%d.ras", 1));

	count = 0;
	list = list_create();

	factor = 1;
	level = 1;
	size = MAX(input.width, input.height);
	while (size > 4*S) {
		printf("Processing level %d...\n", factor);
		count += getstrokes(&input, &area, &list, S, factor, level);
		size /= 2;
		factor *= 2;
		level++;
	}

	list_mergesort(&list);

	image_create(input.width, input.height, IMAGE_RGB, &output);
	mark = bench_mark();
	printf("Creating painterly image.\n");
	gauge_setmax(count);
	gauge_start();
	{
		int xc, yc, w, l;
		int level;
		uchar rgb[3];
		float theta;
		int done = 1;
		p_list p = list;
		while ((p = list_enum(p, &xc, &yc, &w, &l, &theta, rgb, &level))) {
			t_image scaled;
			t_image rotated;
			if (done % (count/10) == 0)
				save_rgb(&output, string("partial-%d.ras", 10*done/count));
			warp_scale(&stroke, l, w, &scaled);
			warp_rotatemono(&scaled, theta, &rotated);
			image_blendrgb(&output, &rotated, rgb, xc, yc);
			gauge_increase(1);
			gauge_print();
			done++;
		}
	}
	gauge_end();
	printf("Done in %.2fs.\n", bench_gettime(mark));
	save_rgb(&output, argv[3]);

    bench_close();

    return 0;
}

