/*--------------------------------------------------------------------*\
* Bilinear scale and rotation
* 3D group (Danilo, Diego and Diogo)
\*--------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>
#include <limits.h>

#include "image.h"
#include "warp.h"

/*--------------------------------------------------------------------*\
* Internal function prototypes
\*--------------------------------------------------------------------*/
static void hscale_mono(p_image in, int ow, p_image out, uchar *temp);
static void hscale_rgb(p_image in, int ow, p_image out, uchar *temp);
static void vscale_mono(p_image in, int oh, p_image out, uchar *temp);
static void vscale_rgb(p_image in, int oh, p_image out, uchar *temp);
static void hshear_mono(p_image in, float dx, p_image out, uchar *temp);
static void vshear_mono(p_image in, float dy, p_image out, uchar *temp);
static void rotate90_mono(p_image in, p_image out, uchar *temp);
static void rotate180_mono(p_image in, p_image out, uchar *temp);
static void rotate270_mono(p_image in, p_image out, uchar *temp);
static void rotate_mono(p_image in, float a, p_image out);

/*--------------------------------------------------------------------*\
* Global variables 
\*--------------------------------------------------------------------*/
static uchar *start = NULL;
static uchar *middle = NULL;
static uchar *end = NULL;
static int length = 0;

int warp_open(int w_max, int h_max)
{
	if (start || end || middle)
		return 0;
	assert(w_max > 0 && h_max > 0);
	if (w_max <= 0 || h_max <= 0)
		return 0;
	length = h_max * w_max;
	start = (uchar *) malloc(length * 3);
	middle = (uchar *) malloc(length * 3);
	end = (uchar *) malloc(length * 3);
	if (!start || !end || !middle) {
		length = 0;
		if (start) free(start); start = NULL;
		if (end) free(end); end = NULL;
		if (middle) free(middle); middle = NULL;
		return 0;
	}
	return 1;
}

int warp_scale(const p_image in, int w, int h, p_image out)
{
	t_image temp;
	assert(in && out);
	if (!in || !out)
		return 0;
	assert(w*h <= length);
	if (w*h > length)
		return 0;
  switch (in->type) {
  case IMAGE_MONO:
    hscale_mono(in, w, &temp, start);
	  vscale_mono(&temp, h, out, end);
    break;
  case IMAGE_RGB:
  	hscale_rgb(in, w, &temp, start);
	  vscale_rgb(&temp, h, out, end);
    break;
  default:
    assert(0);
    break;
  }
  return 1;
}

#define PI 3.141592653589f
int warp_rotatemono(const p_image in, float a, p_image out)
{
	t_image temp;
	assert(in && out);
	if (!in || !out)
		return 0;
	
	a = (float) fmod(a, 2*PI);
	if (a < 0.0f)
		a += 2*PI;

	switch ((int) ((a/(PI/4))) % 8) {
		case 0:
			rotate_mono(in, a, out);
			break;
		case 1:
		case 2:
			rotate90_mono(in, &temp, middle);
			rotate_mono(&temp, a-PI/2, out);
			break;
		case 3:
		case 4:
			rotate180_mono(in, &temp, middle);
			rotate_mono(&temp, a-PI, out);
			break;
		case 5:
		case 6:
			rotate270_mono(in, &temp, middle);
			rotate_mono(&temp, a-3*PI/2, out);
			break;
		case 7:
			rotate_mono(in, a-2*PI, out);
			break;
		default:
			/* bisonho... */
			printf("warp_rotate: a = %f, 4a/PI = %f\n", a, (a/(PI/4)));
			assert(0);
			return 0;
	}

	image_compactmono(out);

	return 1;
}

int warp_close(void)
{
	if (start || end || middle) {
		if (start) free(start); start = NULL;
		if (end) free(end); end = NULL;
		if (middle) free(middle); middle = NULL;
		length = 0;
		return 1;
	}
	return 0;
}

static void hshear_mono(p_image in, float dx, p_image out, uchar *temp)
{
	int x, y;
	int w = in->width + (int) ceil(in->height*fabs(dx));
	float r, f = dx >= 0 ? 0 : -dx * in->height;
	uchar *output = temp;
	uchar *input = in->buffer;
	long skip = (in->row - in->width)*in->type;
	int i;

	out->type = in->type;
	out->buffer = temp;
	out->reference = 1;
	out->height = in->height;
	out->width = w;
	out->row = w;
	image_clear(out, 0);

	for (y = 0; y < in->height; y++) {
		i = (int) floor(f);
		r = f - i;
		for (x = 0; x < in->width; x++) {
			output[i+x] += (uchar) ((1-r)*(*input));
			output[i+x+1] += (uchar) (r*(*input));
			input++;
		}
		input += skip;
		output += w;
		f += dx;
	}
}

static void vshear_mono(p_image in, float dy, p_image out, uchar *temp)
{
	int x, y;
	int h = in->height + (int) ceil(in->width*fabs(dy));
	int w = in->width;
	float r, f = dy >= 0 ? 0 : -dy * in->width;
	int i;
	uchar *input, *output;

	out->type = in->type;
	out->buffer = temp;
	out->reference = 1;
	out->height = h;
	out->width = in->width;
	out->row = in->width;
	image_clear(out, 0);

	for (x = 0; x < in->width; x++) {
		i = (int) floor(f);
		r = f - i;
		output = temp + i*w;
		input = in->buffer;
		for (y = 0; y < in->height; y++) {
			output[x] += (uchar) ((1-r)*(input[x]));
			output[w+x] += (uchar) (r*(input[x]));
			input += in->row;
			output += w;
		}
		f += dy;
	}
}

static void hscale_mono(p_image in, int ow, p_image out, uchar *temp)
{
	int u, x, y;
	long int area; 
	long int res, acc;
	long int p, q;
	uchar *input = in->buffer;
	uchar *output = temp;
	int iw = in->width;
	area = (ow * iw);
	for (y = 0; y < in->height; y++) {
		q = iw;
		p = ow;
		acc = 0;
		x = u = 0;
		while (x < ow) {
			if (u+1 < iw) 
				res = p * input[u] + (ow - p) * input[u+1];
			else
				res = ow * input[u];
			if (p < q) {
				acc += res * p;
				q -= p;
				p = ow;
				u++;
			} else {
				acc += res * q;
				output[x] = (unsigned char ) (acc / area);
				acc = 0;
				p -= q;
				q = iw;
				x++;
			}
		}
		output += ow;
		input += in->row;
	}

	out->type = in->type;
	out->buffer = temp;
	out->reference = 1;
	out->width = ow;
	out->height = in->height;
	out->row = ow;
}

static void hscale_rgb(p_image in, int ow, p_image out, uchar *temp)
{
	int u, x, y;
	long int area; 
	long int res_r, acc_r;
	long int res_g, acc_g;
	long int res_b, acc_b;
	long int p, q;
	uchar *input = in->buffer;
	uchar *output = temp;
	int iw = in->width;
	area = (ow * iw);
	for (y = 0; y < in->height; y++) {
		q = iw;
		p = ow;
		acc_r = 0;
		acc_g = 0;
		acc_b = 0;
		x = u = 0;
		while (x < (3*ow)) {
			if ((u+5) < (3*iw))  {
				res_r = p * input[u] + (ow - p) * input[u+3];
				res_g = p * input[u+1] + (ow - p) * input[u+4];
				res_b = p * input[u+2] + (ow - p) * input[u+5];
			} else {
				res_r = ow * input[u];
				res_g = ow * input[u+1];
				res_b = ow * input[u+2];
			}
			if (p < q) {
				acc_r += res_r * p;
				acc_g += res_g * p;
				acc_b += res_b * p;
				q -= p;
				p = ow;
				u += 3;
			} else {
				acc_r += res_r * q;
				acc_g += res_g * q;
				acc_b += res_b * q;
				output[x] = (unsigned char ) (acc_r / area);
				output[x+1] = (unsigned char ) (acc_g / area);
				output[x+2] = (unsigned char ) (acc_b / area);
				acc_r = 0;
				acc_g = 0;
				acc_b = 0;
				p -= q;
				q = iw;
				x += 3;
			}
		}
		output += ow*3;
		input += in->row*3;
	}

	out->type = in->type;
	out->buffer = temp;
	out->reference = 1;
	out->width = ow;
	out->height = in->height;
	out->row = ow;
}

static void vscale_mono(p_image in, int oh, p_image out, uchar *temp)
{
	int x, y, v;
	long int area; 
	long int res, acc;
	long int p, q;
	uchar *input, *output;
	int ih = in->height;
	long row = in->row*in->type;
	area = (ih * oh);
	for (x = 0; x < in->width; x++) {
		q = ih;
		p = oh;
		acc = 0;
		y = v = 0;
		input = in->buffer;
		output = temp;
		while (y < oh) {
			if (v+1 < ih)
				res = p * input[x] + (oh - p) * input[x+row];
			else
				res = oh * input[x];
			if (p < q) {
				acc += res * p;
				q -= p;
				p = oh;
				input += in->row;
				v++;
			} else {
				acc += res * q;
				output[x] = (unsigned char ) (acc / area);
				acc = 0;
				p -= q;
				q = ih;
				output += in->width;
				y++;
			}
		}
	}

	out->type = in->type;
	out->buffer = temp;
	out->reference = 1;
	out->width = in->width;
	out->height = oh;
	out->row = in->width;
}

static void vscale_rgb(p_image in, int oh, p_image out, uchar *temp)
{
	int x, y, v;
	long int area; 
	long int res_r, acc_r;
	long int res_g, acc_g;
	long int res_b, acc_b;
	long int p, q;
	uchar *input, *output;
	int ih = in->height;
	long row = in->row*in->type;
	area = (ih * oh);
	for (x = 0; x < 3*in->width; x+=3) {
		q = ih;
		p = oh;
		acc_r = 0;
		acc_g = 0;
		acc_b = 0;
		y = v = 0;
		input = in->buffer;
		output = temp;
		while (y < oh) {
			if (v+1 < ih) {
				res_r = p * input[x] + (oh - p) * input[x+row];
				res_g = p * input[x+1] + (oh - p) * input[x+1+row];
				res_b = p * input[x+2] + (oh - p) * input[x+2+row];
			} else {
				res_r = oh * input[x];
				res_g = oh * input[x+1];
				res_b = oh * input[x+2];
			}
			if (p < q) {
				acc_r += res_r * p;
				acc_g += res_g * p;
				acc_b += res_b * p;
				q -= p;
				p = oh;
				input += in->row*3;
				v++;
			} else {
				acc_r += res_r * q;
				acc_g += res_g * q;
				acc_b += res_b * q;
				output[x] = (unsigned char ) (acc_r / area);
				output[x+1] = (unsigned char ) (acc_g / area);
				output[x+2] = (unsigned char ) (acc_b / area);
				acc_r = 0;
				acc_g = 0;
				acc_b = 0;
				p -= q;
				q = ih;
				output += in->width*3;
				y++;
			}
		}
	}

	out->type = in->type;
	out->buffer = temp;
	out->reference = 1;
	out->width = in->width;
	out->height = oh;
	out->row = in->width;
}

static void rotate90_mono(p_image in, p_image out, uchar *temp)
{
	int x, y;	
	uchar *input, *output;
	long skip = in->row - in->width;

	out->buffer = temp;
	out->type = in->type;
	out->reference = 1;
	out->width = in->height;
	out->height = in->width;
	out->row = in->height;

	input = in->buffer;
	for (y = 0; y < in->height; y++) {
		output = out->buffer + in->height - y - 1;
		for (x = 0; x < in->width; x++) {
			*output = *input;
			output += out->row; 
			input++;
		}
		input += skip;
	}
}

static void rotate180_mono(p_image in, p_image out, uchar *temp)
{
	int x, y;	
	uchar *input, *output;
	long skip = in->row - in->width;

	out->buffer = temp;
	out->type = in->type;
	out->reference = 1;
	out->width = in->width;
	out->height = in->height;
	out->row = in->width;

	input = in->buffer;
	output = out->buffer + in->height*in->width - 1;
	for (y = 0; y < in->height; y++) {
		for (x = 0; x < in->width; x++)
			*output-- = *input++;
		input += skip;
	}
}

static void rotate270_mono(p_image in, p_image out, uchar *temp)
{
	int x, y;	
	uchar *input, *output;
	long skip = in->row - in->width;
	long whole = in->height*(in->width-1);

	out->buffer = temp;
	out->type = in->type;
	out->reference = 1;
	out->width = in->height;
	out->height = in->width;
	out->row = in->height;

	input = in->buffer;
	for (y = 0; y < in->height; y++) {
		output = out->buffer + whole + y;
		for (x = 0; x < in->width; x++) {
			*output = *input;
			output -= out->row; 
			input++;
		}
		input += skip;
	}
}

static void rotate_mono(p_image in, float a, p_image out)
{
	t_image t1, t2;
    hshear_mono(in, -(float)(tan(a/2)), &t1, start);
    vshear_mono(&t1, (float)sin(a), &t2, middle);
    hshear_mono(&t2, -(float)(tan(a/2)), out, end);
}
